#include "GameFileManager.hpp"
#include "GameManager.hpp"
#include "Isomot.hpp"
#include "MapManager.hpp"
#include "BonusManager.hpp"
#include "Mediator.hpp"
#include "PlayerItem.hpp"
#include "Room.hpp"

namespace isomot
{

//--------------------------------------------------------------------------------------------------
GameFileManager::GameFileManager(const GameManager* gameManager, const Isomot* isomot)
: roomId(std::string()),
  label(0),
  x(0),
  y(0),
  z(0),
  direction(NoDirection),
  gameManager(const_cast<GameManager*>(gameManager)),
  isomot(const_cast<Isomot*>(isomot))
{
  
}

//--------------------------------------------------------------------------------------------------
GameFileManager::~GameFileManager()
{
  
}

//--------------------------------------------------------------------------------------------------
void GameFileManager::assignFishData(const std::string& roomId, short label, int x, int y, int z, const Direction& direction)
{
  this->roomId = roomId;
  this->label = label;
  this->x = x;
  this->y = y;
  this->z = z;
  this->direction = direction;
}

//--------------------------------------------------------------------------------------------------
void GameFileManager::loadGame(const std::string& fileName)
{ 
   try
   {
      boost::property_tree::ptree xmlTree;
      boost::property_tree::xml_parser::read_xml(fileName, xmlTree);
    
      // Establece las salas que ya han sido visitadas.
      this->isomot->getMapManager()->load(xmlTree.get_child("savegame.exploredRooms"));
    
      // Asigna los bonus que no deben aparecer en esta partida.
      BonusManager::getInstance()->load(xmlTree.get_child("savegame.bonus"));

      if(xmlTree.get<bool>("savegame.freeByblos"))
      {
         this->gameManager->liberatePlanet(Byblos, false);
      }
      if(xmlTree.get<bool>("savegame.freeEgyptus"))
      {
         this->gameManager->liberatePlanet(Egyptus, false);
      }
      if(xmlTree.get<bool>("savegame.freePenitentiary"))
      {
         this->gameManager->liberatePlanet(Penitentiary, false);
      }
      if(xmlTree.get<bool>("savegame.freeSafari"))
      {
         this->gameManager->liberatePlanet(Safari, false);
      }
      if(xmlTree.get<bool>("savegame.freeBlacktooth"))
      {
         this->gameManager->liberatePlanet(Blacktooth, false);
      }
    
      // Asigna el estado de los jugadores.
      this->assignPlayerStatus(xmlTree.get_child("savegame.players"));
      // Crea las salas iniciales y la posición de los jugadores.
      this->isomot->start(xmlTree.get_child("savegame.players"));
   }
   catch (const std::runtime_error& e)
   {
      std::cerr << e.what() << std::endl;
   }
}

//--------------------------------------------------------------------------------------------------
void GameFileManager::saveGame(const std::string& fileName)
{
  try
  {
     boost::property_tree::ptree xmlTree;
     xmlTree.put("savegame.<xmlattr>.xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
     xmlTree.put("savegame.<xmlattr>.xsi:noNamespaceSchemaLocation", "savegame.xsd");

     // Se almacenan las salas visitadas.
     this->isomot->getMapManager()->save(xmlTree);
    
     // Planetas liberados.
     xmlTree.put("savegame.freeByblos", this->gameManager->isFreePlanet(Byblos));
     xmlTree.put("savegame.freeEgyptus", this->gameManager->isFreePlanet(Egyptus));
     xmlTree.put("savegame.freePenitentiary", this->gameManager->isFreePlanet(Penitentiary));
     xmlTree.put("savegame.freeSafari", this->gameManager->isFreePlanet(Safari));
     xmlTree.put("savegame.freeBlacktooth", this->gameManager->isFreePlanet(Blacktooth));

     // Se almacenan los bonus que se hallan cogido.
     BonusManager::getInstance()->save(xmlTree);
    
     // Se almacena el estado de los jugadores.
     // El jugador activo.
     PlayerId activePlayerId = PlayerId(this->label);
    
     // Número de vidas.
     unsigned short lives = 0;
     switch(activePlayerId)
     {
        case Head:
        case Heels:
           lives = this->gameManager->getLives(activePlayerId);
           break;
        
        case HeadAndHeels:
           lives = this->gameManager->getLives(Head) * 100 + this->gameManager->getLives(Heels);
           break;
        
        default:
           ;
     }
     // Posesión de objetos.
     std::vector<short> tools = this->gameManager->hasTool(activePlayerId);

     // Almacenamiento de todos los datos.
     boost::property_tree::ptree players;

     boost::property_tree::ptree activePlayerTree;
     activePlayerTree.put("<xmlattr>.label", this->label);
     activePlayerTree.put("active", true);
     activePlayerTree.put("roomFilename", this->roomId);
     activePlayerTree.put("x", this->x);
     activePlayerTree.put("y", this->y);
     activePlayerTree.put("z", this->z);
     activePlayerTree.put("direction", int(this->direction));
     activePlayerTree.put("entry", Wait);
     activePlayerTree.put("lives", lives);
     activePlayerTree.put("hasHorn", activePlayerId == Head || activePlayerId == HeadAndHeels ? std::find(tools.begin(), tools.end(), short(Horn)) != tools.end() : false);
     activePlayerTree.put("hasHandbag", activePlayerId == Heels || activePlayerId == HeadAndHeels ? std::find(tools.begin(), tools.end(), short(Handbag)) != tools.end() : false);
     activePlayerTree.put("ammo", this->gameManager->getDonuts(activePlayerId));
     players.add_child("player", activePlayerTree);
    
     // Es posible que no haya más salas bien porque no haya más jugadores o porque el otro jugador
     // se encuentre en la misma sala que el jugador activo.
     Room* hideRoom = this->isomot->getMapManager()->getHideRoom();
     Room* activeRoom = this->isomot->getMapManager()->getActiveRoom();
     PlayerItem* inactivePlayer = (hideRoom != 0 ? hideRoom->getMediator()->getActivePlayer() : activeRoom->getMediator()->getHidePlayer());
    
     // Si hay algún otro jugador, se almacenan sus datos.
     if(inactivePlayer != 0)
     {
        PlayerId inactivePlayerId = PlayerId(inactivePlayer->getLabel());
        std::vector<short> tools = this->gameManager->hasTool(inactivePlayerId);
        PlayerStartPosition* startPosition = this->isomot->getMapManager()->findPlayerPosition(hideRoom != 0 ? hideRoom->getIdentifier() : activeRoom->getIdentifier(), PlayerId(inactivePlayer->getLabel()));
        if(startPosition == 0)
        {
           startPosition = this->isomot->getMapManager()->findPlayerPosition(activeRoom->getIdentifier(), HeadAndHeels);
        }

        boost::property_tree::ptree inactivePlayerTree;
        inactivePlayerTree.put("<xmlattr>.label", inactivePlayer->getLabel());
        inactivePlayerTree.put("active", false);
        inactivePlayerTree.put("roomFilename", hideRoom != 0 ? hideRoom->getIdentifier() : activeRoom->getIdentifier());
        inactivePlayerTree.put("x", startPosition->getX());
        inactivePlayerTree.put("y", startPosition->getY());
        inactivePlayerTree.put("z", startPosition->getZ());
        inactivePlayerTree.put("direction", int(inactivePlayer->getDirection()));
        inactivePlayerTree.put("entry", startPosition->getEntry());
        inactivePlayerTree.put("lives", this->gameManager->getLives(inactivePlayerId));
        inactivePlayerTree.put("hasHorn", inactivePlayerId == Head || inactivePlayerId == HeadAndHeels ? std::find(tools.begin(), tools.end(), short(Horn)) != tools.end() : false);
        inactivePlayerTree.put("hasHandbag", inactivePlayerId == Heels || inactivePlayerId == HeadAndHeels ? std::find(tools.begin(), tools.end(), short(Handbag)) != tools.end() : false);
        inactivePlayerTree.put("ammo", this->gameManager->getDonuts(inactivePlayerId));
        players.add_child("player", inactivePlayerTree);

        xmlTree.add_child("savegame.players", players);
     }
    
     // Grabación del archivo.
     auto settings = boost::property_tree::xml_parser::xml_writer_make_settings<boost::property_tree::ptree::key_type>(' ', 3);
     boost::property_tree::xml_parser::write_xml(fileName, xmlTree, std::locale(), settings);
  }
  catch (const std::runtime_error& e)
  {
     std::cerr << e.what() << std::endl;
  }
}

void GameFileManager::assignPlayerStatus(const boost::property_tree::ptree& playerSequence)
{
   auto&& playerIter = playerSequence.equal_range("player");
   for (auto&& i = playerIter.first; i != playerIter.second; ++i)
   {
      unsigned char headLives = 0;
      unsigned char heelsLives = 0;

      // Si se va a crear al jugador compuesto se obtienen las vidas de los jugadores simples.
      // Se utiliza la fórmula: Vidas H&H = Vidas Head * 100 + Vidas Heels.
      if (i->second.get<short>("<xmlattr>.label") == HeadAndHeels)
      {
         unsigned short lives = i->second.get<short>("lives");

         while (lives > 100)
         {
            lives -= 100;
            headLives++;
         }
         heelsLives = static_cast<unsigned char>(lives);
      }

      switch (i->second.get<short>("<xmlattr>.label"))
      {
         case Head:
            this->gameManager->setHeadLives(i->second.get<short>("lives"));
            this->gameManager->setHorn(i->second.get<bool>("hasHorn"));
            this->gameManager->setDonuts(i->second.get<unsigned short>("ammo"));
            this->gameManager->setHighSpeed(0);
            this->gameManager->setHeadShield(0);
            break;

         case Heels:
            this->gameManager->setHeelsLives(i->second.get<short>("lives"));
            this->gameManager->setHandbag(i->second.get<bool>("hasHandbag"));
            this->gameManager->setHighJumps(0);
            this->gameManager->setHeelsShield(0);
            break;

         case HeadAndHeels:

            this->gameManager->setHeadLives(headLives);
            this->gameManager->setHorn(i->second.get<bool>("hasHorn"));
            this->gameManager->setDonuts(i->second.get<unsigned short>("ammo"));
            this->gameManager->setHeelsLives(heelsLives);
            this->gameManager->setHandbag(i->second.get<bool>("hasHandbag"));
            this->gameManager->setHighSpeed(0);
            this->gameManager->setHighJumps(0);
            this->gameManager->setHeadShield(0);
            this->gameManager->setHeelsShield(0);
            break;
      }
   }
}

}
