#include "ItemDataManager.hpp"
#include "Ism.hpp"

namespace isomot
{

ItemDataManager::ItemDataManager(const std::string& fileName)
   :
   fileName{fileName}
{

}

ItemDataManager::~ItemDataManager()
{
   std::for_each(this->itemDataVector.begin(), this->itemDataVector.end(), destroyItemData);
}

void ItemDataManager::load()
{ 
   // Carga el archivo XML especificado y almacena los datos XML en la lista.
   try
   {
      boost::property_tree::ptree xmlTree;
      boost::property_tree::xml_parser::read_xml(isomot::sharePath() + fileName, xmlTree);

      auto&& itemIter = xmlTree.get_child("items").equal_range("item");
      for (auto&& i = itemIter.first; i != itemIter.second; ++i)
      {
         processItem(i->second);
      }
   }
   catch (const std::runtime_error& e)
   {
      std::cerr << e.what() << std::endl;
   }
}

void ItemDataManager::processItem(const boost::property_tree::ptree& item)
{
   ItemData itemData;

   // Identificador único y exclusivo del elemento.
   itemData.label = item.get<short>("<xmlattr>.label");
   // Número de fotogramas en función de la dirección del elemento.
   itemData.directionFrames = item.get<unsigned char>("directionFrames");
   // Elemento ofensivo o inofensivo.
   itemData.mortal = item.get<bool>("mortal");
   // Cada cuántos milisegundos cae el elemento.
   itemData.weight = item.get<double>("weight");
   // Cada cuántos milisegundos cambian los fotogramas de la secuencia de animación.
   itemData.framesDelay = item.get<double>("framesDelay");
   // Cada cuántos milisegundos se desplaza el elemento una unidad isométrica.
   itemData.speed = item.get<double>("speed");
   // Ruta a los gráficos del elemento.
   itemData.bitmap = isomot::sharePath() + item.get<std::string>("bitmap.<xmlattr>.file");
   // Anchura en píxeles de un fotograma.
   itemData.frameWidth = item.get<int>("bitmap.frameWidth");
   // Altura en píxeles de un fotograma.
   itemData.frameHeight = item.get<int>("bitmap.frameHeight");

   // Almacena las sombras del elemento.
   processShadow(item, itemData);

   // Sólo unos pocos elementos tienen fotogramas extra.
   itemData.extraFrames = item.get("extraFrames", 0);

   // Secuencia de animación.
   auto&& frameIter = item.equal_range( "frame" );
   for (auto&& i = frameIter.first; i != frameIter.second; ++i)
   {
      itemData.frames.push_back(std::stoi(i->second.data()));
   }

   // ¿Es una puerta?
   bool isDoor = !item.get("door", "").empty();

   // Se almacenan los fotogramas en el vector motion siempre que el elemento no sea una puerta.
   if (!isDoor)
   {
      createBitmapFrames(itemData);

      // Anchura en el eje X en unidades isométricas del elemento.
      itemData.widthX = item.get<int>("widthX");
      // Anchura en el eje Y en unidades isométricas del elemento.
      itemData.widthY = item.get<int>("widthY");
      // Anchura en el eje Z en unidades isométricas del elemento; su altura.
      itemData.height = item.get<int>("height");

      // Los datos del elemento simple se incorporan a la lista.
      this->itemDataVector.push_back(itemData);
   }
   else
   {
      processDoor(item, itemData);
   }
}

void ItemDataManager::processShadow(const boost::property_tree::ptree& item, ItemData& itemData)
{
   // El elemento puede no tener sombra.
   if (std::string shadowFile{item.get("shadow.<xmlattr>.file", "")}; !shadowFile.empty())
   {
      // Archivo de la sombra del elemento.
      itemData.shadow = isomot::sharePath() + shadowFile;
      // Anchura en píxeles de un fotograma de sombra.
      itemData.shadowWidth = item.get<int>("shadow.shadowWidth");
      // Altura en píxeles de un fotograma de sombra.
      itemData.shadowHeight = item.get<int>("shadow.shadowHeight");
      // Se almacenan los fotogramas en el vector shadows.
      createShadowFrames(itemData);
   }
}

void ItemDataManager::processDoor(const boost::property_tree::ptree& item, const ItemData& itemData)
{
   // Si el elemento es una puerta, tiene tres parámetros que definen sus dimensiones.
   // Una puerta en realidad son tres elementos distintos: dintel, jamba izquierda y jamba derecha.
   ItemData lintel(itemData);
   ItemData leftJamb(itemData);
   ItemData rightJamb(itemData);

   // Dimensiones de las partes de una puerta.
   DoorMeasures dm;

   // Anchura espacial en el eje X de las tres partes de la puerta.
   // Se almacenan en el siguiente orden: jamba izquierda, jamba derecha y dintel.
   auto&& widthXIter = item.equal_range( "widthX" );
   for (auto&& i = widthXIter.first; i != widthXIter.second; ++i)
   {
      if (dm.leftJambWidthX == 0)
      {
        leftJamb.widthX = dm.leftJambWidthX = std::stoi(i->second.data());
      }
      else if (dm.rightJambWidthX == 0)
      {
        rightJamb.widthX = dm.rightJambWidthX = std::stoi(i->second.data());
      }
      else if (dm.lintelWidthX == 0)
      {
        lintel.widthX = dm.lintelWidthX = std::stoi(i->second.data());
      }
   }
        
   // Anchura espacial en el eje Y de las tres partes de la puerta.
   // Se almacenan en el siguiente orden: jamba izquierda, jamba derecha y dintel.
   auto&& widthYIter = item.equal_range( "widthY" );
   for (auto&& i = widthYIter.first; i != widthYIter.second; ++i)
   {
      if (dm.leftJambWidthY == 0)
      {
         leftJamb.widthY = dm.leftJambWidthY = std::stoi(i->second.data());
      }
      else if (dm.rightJambWidthY == 0)
      {
         rightJamb.widthY = dm.rightJambWidthY = std::stoi(i->second.data());
      }
      else if (dm.lintelWidthY == 0)
      {
         lintel.widthY = dm.lintelWidthY = std::stoi(i->second.data());
      }
   }
        
   // Altura espacial de las tres partes de la puerta.
   // Se almacenan en el siguiente orden: jamba izquierda, jamba derecha y dintel.
   auto&& heightIter = item.equal_range( "height" );
   for (auto&& i = heightIter.first; i != heightIter.second; ++i)
   {
      if (dm.leftJambHeight == 0)
      {
         leftJamb.height = dm.leftJambHeight = std::stoi(i->second.data());
      }
      else if (dm.rightJambHeight == 0)
      {
         rightJamb.height = dm.rightJambHeight = std::stoi(i->second.data());
      }
      else if (dm.lintelHeight == 0)
      {
         lintel.height = dm.lintelHeight = std::stoi(i->second.data());
      }
   }

   // Se cargan el gráfico de la puerta al completo y se obtienen los gráficos de sus tres partes.
   BITMAP* gfx = load_bitmap(itemData.bitmap.c_str(), 0);
   if (gfx == nullptr)
   {
      throw std::logic_error{"Bitmap " + itemData.bitmap + " not found"};
   }
        
   std::string door{item.get<std::string>("door")};

   // Creación de la jamba izquierda.
   BITMAP* left = cutOutLeftJamb(gfx, dm, door);
   leftJamb.motion.push_back(left);
   this->itemDataVector.push_back(leftJamb);
        
   // Creación de la jamba derecha.
   BITMAP* right = cutOutRightJamb(gfx, dm, door);
   rightJamb.label += 1;
   rightJamb.motion.push_back(right);
   this->itemDataVector.push_back(rightJamb);
        
   // Creación del dintel.
   BITMAP* top = cutOutLintel(gfx, dm, door);
   lintel.label += 2;
   lintel.motion.push_back(top);
   this->itemDataVector.push_back(lintel);

   // La imagen original no se volverá a utilizar.
   destroy_bitmap(gfx);
}

ItemData* ItemDataManager::find(short label)
{
   auto&& i = std::find_if(this->itemDataVector.begin(), this->itemDataVector.end(),
                           [label](const ItemData& itemData) {
                              return itemData.label == label;
                           });

   return (i != this->itemDataVector.end() ? &(*i) : nullptr);
}

void ItemDataManager::createBitmapFrames(ItemData& itemData)
{ 
   // Datos necesarios.
   if (itemData.bitmap.empty() || itemData.frameWidth == 0 || itemData.frameHeight == 0)
   {
      throw std::logic_error{"Insufficient data for ItemDataManager::createBitmapFrames"};
   }
    
   // Se cargan los gráficos del elemento y si es necesario se crean los distintos fotogramas.
   BITMAP* gfx = load_bitmap(itemData.bitmap.c_str(), 0);
   if (gfx == nullptr)
   {
      throw std::logic_error{"Bitmap " + itemData.bitmap + " not found"};
   }

   // Trocea la imagen en los distintos fotogramas y se almacenan en el vector.
   for (int y = 0; y < gfx->h / itemData.frameHeight; y++)
   {
      for (int x = 0; x < gfx->w / itemData.frameWidth; x++)
      {
         BITMAP* frame = create_bitmap_ex(32, itemData.frameWidth, itemData.frameHeight);
         blit(gfx, frame, x * itemData.frameWidth, y * itemData.frameHeight,
                   0, 0,
                   frame->w, frame->h);
         itemData.motion.push_back(frame);
      }
   }
    
   // La imagen original no se volverá a utilizar.
   destroy_bitmap(gfx);
}

void ItemDataManager::createShadowFrames(ItemData& itemData)
{  
   // Datos necesarios.
   if (itemData.shadow.empty() || itemData.shadowWidth == 0 || itemData.shadowHeight == 0)
   {
      throw std::logic_error{"Insufficient data for ItemDataManager::createShadowFrames"};
   }
    
   // Se cargan los gráficos del elemento y si es necesario se crean los distintos fotogramas.
   BITMAP* gfx = load_bitmap(itemData.shadow.c_str(), 0);
   if (gfx == nullptr)
   {
      throw std::logic_error{"Bitmap " + itemData.shadow + " not found"};
   }

   // Trocea la imagen en los distintos fotogramas y se almacenan en el vector.
   for (int y = 0; y < gfx->h / itemData.shadowHeight; y++)
   {
      for(int x = 0; x < gfx->w / itemData.shadowWidth; x++)
      {
         BITMAP* frame = create_bitmap_ex(32, itemData.shadowWidth, itemData.shadowHeight);
         blit(gfx, frame, x * itemData.shadowWidth, y * itemData.shadowHeight,
                   0, 0,
                   frame->w, frame->h);
         itemData.shadows.push_back(frame);
      }
   }
    
   // La imagen original no se volverá a utilizar.
   destroy_bitmap(gfx);
}

BITMAP* ItemDataManager::cutOutLintel(BITMAP* door, const DoorMeasures& dm, const std::string& type)
{
   bool ns{type == "north" || type == "south"};
  
   // Parte superior de la puerta.
   BITMAP* top = create_bitmap_ex(32, (dm.lintelWidthX << 1) + (dm.lintelWidthY << 1),
                                       dm.lintelHeight + dm.lintelWidthY + dm.lintelWidthX);
   // El color de fondo es magenta.
   clear_to_color(top, makecol(255, 0, 255));

   // Copia la zona a recortar si la puerta está orientada al norte o al sur.
   if (ns)
   {
      // Copia la zona rectangular.
      blit(door, top, 0, 0, 0, 0, top->w, dm.lintelHeight + dm.lintelWidthX);

      int delta = top->w - 1;
      int noPixelIndex = top->w - ((dm.rightJambWidthX + dm.rightJambWidthY) << 1) + 1;
      int yStart = noPixelIndex;
      int yEnd = noPixelIndex - 1;

      acquire_bitmap(top);

      for (int x = dm.lintelHeight + dm.lintelWidthX; x < top->h; x++)
      {
         for (int y = delta; y >= 0; y--)
         {
            if (x != dm.lintelHeight + dm.lintelWidthX && y == noPixelIndex)
            {
               if (noPixelIndex > yEnd)
               {
                  noPixelIndex--;
               }
               else
               {
                  yStart += 2;
                  noPixelIndex = yStart;
               }
            }
            else if (x < dm.lintelHeight + (dm.lintelWidthX << 1) || y < yEnd)
            {
               ((int*) top->line[x])[y] = ((int*) door->line[x])[y];
            }
         }

         delta -= 2;
      }

      release_bitmap(top);
   }
   // Copia la zona a recortar si la puerta está orientada al este o al oeste.
   else
   {
      // Copia la zona rectangular.
      blit(door, top, 0, 0, 0, 0, top->w, dm.lintelHeight + dm.lintelWidthY);

      int delta = 0;
      int noPixelIndex = ((dm.leftJambWidthX + dm.leftJambWidthY) << 1) - 2;
      int yStart = noPixelIndex;
      int yEnd = noPixelIndex + 1;

      acquire_bitmap(top);

      for (int x = dm.lintelHeight + dm.lintelWidthY; x < top->h; x++)
      {
         for (int y = delta; y < top->w; y++)
         {
            if (x != dm.lintelHeight + dm.lintelWidthY && y == noPixelIndex)
            {
               if (noPixelIndex < yEnd)
               {
                  noPixelIndex++;
               }
               else
               {
                  yStart -= 2;
                  noPixelIndex = yStart;
               }
            }
            else if (x < dm.lintelHeight + (dm.lintelWidthY << 1) || y > yEnd)
            {
               ((int*) top->line[x])[y] = ((int*) door->line[x])[y];
            }
         }

         delta += 2;
      }

      release_bitmap(top);
   }
  
   return top;
}

BITMAP* ItemDataManager::cutOutLeftJamb(BITMAP* door, const DoorMeasures& dm, const std::string& type)
{
   bool ns{type == "north" || type == "south"};
   int fixWidth = (ns ? 7 : 0);
   int fixY = (ns ? -1 : 0);
  
   BITMAP* left = create_bitmap_ex(32, (dm.leftJambWidthX << 1) + fixWidth + (dm.leftJambWidthY << 1),
                                        dm.leftJambHeight + dm.leftJambWidthY + dm.leftJambWidthX);
  
   // El color de fondo es magenta.
   clear_to_color(left, makecol(255, 0, 255));
   // Copia la zona rectangular.
   blit(door, left,
        fixY, dm.lintelHeight + dm.lintelWidthY - dm.leftJambWidthY + fixY,
        0, 0,
        left->w, left->h);
          
   return left;
}

BITMAP* ItemDataManager::cutOutRightJamb(BITMAP* door, const DoorMeasures& dm, const std::string& type)
{
   bool ns{type == "north" || type == "south"};
   int fixWidth = (ns ? 0 : 7);
   int fixY = (ns ? 0 : -2);
  
   BITMAP* right = create_bitmap_ex(32, (dm.rightJambWidthX << 1) + fixWidth + (dm.rightJambWidthY << 1),
                                         dm.rightJambHeight + dm.rightJambWidthY + dm.rightJambWidthX);
  
   // El color de fondo es magenta.
   clear_to_color(right, makecol(255, 0, 255));
   // Copia la zona rectangular.
   blit(door, right,
        door->w - right->w,
        dm.lintelHeight + dm.lintelWidthX - dm.rightJambWidthY + fixY,
        0, 0,
        right->w, right->h);
  
   return right;
}

void ItemDataManager::destroyItemData(ItemData& itemData)
{
   itemData.bitmap.clear();
   itemData.shadow.clear();
   itemData.frames.clear();
   std::for_each(itemData.motion.begin(), itemData.motion.end(), destroy_bitmap);
   std::for_each(itemData.shadows.begin(), itemData.shadows.end(), destroy_bitmap);
}

}
