#include "RoomBuilder.hpp"
#include "ItemDataManager.hpp"
#include "Room.hpp"
#include "FloorTile.hpp"
#include "Wall.hpp"
#include "GridItem.hpp"
#include "FreeItem.hpp"
#include "PlayerItem.hpp"
#include "Door.hpp"
#include "BonusManager.hpp"
#include "GameManager.hpp"

namespace isomot
{

RoomBuilder::RoomBuilder(ItemDataManager* itemDataManager, const std::string& fileName)
   :
   fileName{fileName},
   itemDataManager{itemDataManager},
   room{},
   directionMap{{"none", NoDirection},
                {"south", South},
                {"west", West},
                {"north", North},
                {"east", East},
                {"northeast", Northeast},
                {"southeast", Southeast},
                {"southwest", Southwest},
                {"northwest", Northwest},
                {"eastnorth", Eastnorth},
                {"eastsouth", Eastsouth},
                {"westnorth", Westnorth},
                {"westsouth", Westsouth}}
{

}

RoomBuilder::RoomBuilder(ItemDataManager* itemDataManager)
   :
   itemDataManager{itemDataManager},
   room{},
   directionMap{{"none", NoDirection},
                {"south", South},
                {"west", West},
                {"north", North},
                {"east", East},
                {"northeast", Northeast},
                {"southeast", Southeast},
                {"southwest", Southwest},
                {"northwest", Northwest},
                {"eastnorth", Eastnorth},
                {"eastsouth", Eastsouth},
                {"westnorth", Westnorth},
                {"westsouth", Westsouth}}
{

}

Room* RoomBuilder::buildRoom()
{
   // Procesa el archivo XML especificado para construir la sala.
   try
   {
      boost::property_tree::ptree xmlTree;
      boost::property_tree::xml_parser::read_xml(fileName, xmlTree);

      std::map<std::string, FloorId> floorMap{{"none", NoFloor},
                                              {"regular", RegularFloor},
                                              {"mortal", MortalFloor}};

      // Crea la sala con los parámetros básicos: identificador, dimensiones y existencia o no de suelo.
      this->room = new Room(xmlTree.get<std::string>("room.name"),
                            xmlTree.get<std::string>("room.scenery"),
                            xmlTree.get<int>("room.xTiles"),
                            xmlTree.get<int>("room.yTiles"),
                            xmlTree.get<int>("room.width"),
                            floorMap[xmlTree.get<std::string>("room.floorType")]);

      if (this->room == 0)
      {
         throw "Room creation error: " + fileName + " has bad atributtes.";
      }

      // Para calcular las coordenadas de pantalla donde se situará el origen del espacio isométrico,
      // se necesita saber si la sala tiene puerta al norte y/o al este.
      bool hasNorthDoor = false;
      bool hasEastDoor = false;

      // Recorre la lista de elementos buscando las puertas.
      auto&& itemIter = xmlTree.get_child("room.items").equal_range("item");
      for (auto&& i = itemIter.first; i != itemIter.second; ++i)
      {
         if (i->second.get<std::string>("type") == "door")
         {
            std::string direction{i->second.get<std::string>("direction")};
            if (direction == "north" || direction == "northeast" || direction == "northwest")
            {
               hasNorthDoor = true;
            }
            else if (direction == "east" || direction == "eastnorth" || direction == "eastsouth")
            {
               hasEastDoor = true;
            }
         }
      }

      // Una vez que se sabe de la existencia o no de las puertas, se calculan las coordenadas.
      room->calculateCoordinates(hasNorthDoor, hasEastDoor,
                                 xmlTree.get<int>("room.deltaX", 0),
                                 xmlTree.get<int>("room.deltaY", 0));

      // Si la sala es triple se almacenan los datos constantes para situar inicialmente la cámara
      // y las coordenadas de la sala donde se realizará el desplazamiento de la cámara.
      auto&& rootNode = xmlTree.get_child("room");
      if (rootNode.find("triple-room-data") != rootNode.not_found())
      {
         auto&& tripleRoomNode = xmlTree.get_child("room.triple-room-data");

         if (tripleRoomNode.find("northeast") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("northeast.x")};
            int y{tripleRoomNode.get<int>("northeast.y")};
            room->addTripleRoomStartPoint(Northeast, x, y);
         }
         if (tripleRoomNode.find("southeast") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("southeast.x")};
            int y{tripleRoomNode.get<int>("southeast.y")};
            room->addTripleRoomStartPoint(Southeast, x, y);
         }
         if (tripleRoomNode.find("northwest") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("northwest.x")};
            int y{tripleRoomNode.get<int>("northwest.y")};
            room->addTripleRoomStartPoint(Northwest, x, y);
         }
         if (tripleRoomNode.find("southwest") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("southwest.x")};
            int y{tripleRoomNode.get<int>("southwest.y")};
            room->addTripleRoomStartPoint(Southwest, x, y);
         }
         if (tripleRoomNode.find("eastnorth") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("eastnorth.x")};
            int y{tripleRoomNode.get<int>("eastnorth.y")};
            room->addTripleRoomStartPoint(Eastnorth, x, y);
         }
         if (tripleRoomNode.find("eastsouth") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("eastsouth.x")};
            int y{tripleRoomNode.get<int>("eastsouth.y")};
            room->addTripleRoomStartPoint(Eastsouth, x, y);
         }
         if (tripleRoomNode.find("westnorth") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("westnorth.x")};
            int y{tripleRoomNode.get<int>("westnorth.y")};
            room->addTripleRoomStartPoint(Westnorth, x, y);
         }
         if (tripleRoomNode.find("westsouth") != tripleRoomNode.not_found())
         {
            int x{tripleRoomNode.get<int>("westsouth.x")};
            int y{tripleRoomNode.get<int>("westsouth.y")};
            room->addTripleRoomStartPoint(Westsouth, x, y);
         }

         int xmin{tripleRoomNode.get<int>("bound-x.minimum")};
         int xmax{tripleRoomNode.get<int>("bound-x.maximum")};
         int ymin{tripleRoomNode.get<int>("bound-y.minimum")};
         int ymax{tripleRoomNode.get<int>("bound-y.maximum")};
         room->assignTripleRoomBounds(xmin, xmax, ymin, ymax);
      }

      // Crea el suelo a base de losetas.
      auto&& tileIter = xmlTree.get_child("room.floor").equal_range("tile");
      for (auto&& i = tileIter.first; i != tileIter.second; ++i)
      {
         FloorTile* floorTile = this->buildFloorTile(i->second);
         room->addComponent(floorTile);
      }

      // Crea las paredes sin puertas.
      if (rootNode.find("walls") != rootNode.not_found())
      {
         auto&& wallIter = xmlTree.get_child("room.walls").equal_range("wall");
         for (auto&& i = wallIter.first; i != wallIter.second; ++i)
         {
            Wall* wall = this->buildWall(i->second);
            room->addComponent(wall);
         }
      }

      // Sitúa los elementos en la sala. Un elemento puede ser: un muro, una puerta, uno rejilla o uno libre.
      auto&& itemIter2 = xmlTree.get_child("room.items").equal_range("item");
      for (auto&& i = itemIter2.first; i != itemIter2.second; ++i)
      {
         std::string type{i->second.get<std::string>("type")};

         // Es una puerta.
         if (type == "door")
         {
            Door* door = this->buildDoor(i->second);

            if (!door)
            {
               std::ostringstream oss;
               oss << "Cannot build door with coordinates "
                   << i->second.get<int>("<xmlattr>.x") << ", "
                   << i->second.get<int>("<xmlattr>.y") << ", "
                   << i->second.get<int>("<xmlattr>.z");
               throw std::runtime_error{oss.str()};
            }

            room->addComponent(door);
         }
         // Es un elemento rejilla.
         else if (type == "griditem")
         {
            GridItem* gridItem = this->buildGridItem(i->second);

            if (gridItem == 0)
            {
               std::ostringstream oss;
               oss << "Cannot build grid-item with coordinates "
                   << i->second.get<int>("<xmlattr>.x") << ", "
                   << i->second.get<int>("<xmlattr>.y") << ", "
                   << i->second.get<int>("<xmlattr>.z");
               throw std::runtime_error{oss.str()};
            }

            room->addComponent(gridItem);
         }
         // Es un elemento libre.
         else if (type == "freeitem")
         {
            FreeItem* freeItem = this->buildFreeItem(i->second);

            if (freeItem == 0)
            {
               // En este caso no se lanza ninguna excepción porque hay
               // elementos -los bonus- que pueden no crearse.
               std::ostringstream oss;
               oss << "Cannot build free-item with coordinates "
                   << i->second.get<int>("<xmlattr>.x") << ", "
                   << i->second.get<int>("<xmlattr>.y") << ", "
                   << i->second.get<int>("<xmlattr>.z");
               std::cout << oss.str() << std::endl;
            }
            else
            {
               room->addComponent(freeItem);
            }
         }
      }

      // Calcula los límites de la sala.
      room->calculateBounds();
   }
   catch (const std::runtime_error& e)
   {
      std::cerr << e.what() << std::endl;
      throw e;
   }

   return this->room;
}

PlayerItem* RoomBuilder::buildPlayerItem(const PlayerId& playerId, const BehaviorId& behaviorId,
                                         int x, int y, int z, const Direction& direction)
{
   return buildPlayerItem(this->room, playerId, behaviorId, x, y, z, direction);
}

PlayerItem* RoomBuilder::buildPlayerItem(Room* room, const PlayerId& playerId,
                                         const BehaviorId& behaviorId, int x, int y, int z,
                                         const Direction& direction, bool hasItem)
{
   PlayerItem* playerItem = 0;
   GameManager* gameManager = GameManager::getInstance();
   PlayerId newPlayerId = playerId;
   BehaviorId newBehaviorId = behaviorId;

   // Si el jugador compuesto se ha quedado sin vidas, se comprueba si alguno de
   // los jugadores conseva alguna, para crearlo en su lugar.
   if (gameManager->getLives(playerId) == 0)
   {
      if (playerId == HeadAndHeels)
      {
         // Jugador Superviviente.
         if (gameManager->getLives(Head) > 0)
         {
            newPlayerId = Head;
            newBehaviorId = HeadBehavior;
         }
         else if (gameManager->getLives(Heels) > 0)
         {
            newPlayerId = Heels;
            newBehaviorId = HeelsBehavior;
         }
         else
         {
            newPlayerId = NoPlayer;
         }
      }
      // Es posible que los dos jugadores se unieran en la sala y se hayan quedado sin vidas.
      else
      {
         if (gameManager->getLives(Head) == 0 && gameManager->getLives(Heels) == 0)
         {
            newPlayerId = NoPlayer;
         }
      }
   }

   // Se buscan los datos del elemento.
   ItemData* itemData = this->itemDataManager->find(short(newPlayerId));

   // Si se han encontrado y al jugador le quedan vidas, se coloca el elemento en la sala.
   if (newPlayerId != NoPlayer && itemData != 0)
   {
      // Para poder crear el jugador le deben quedar vidas.
      if (gameManager->getLives(newPlayerId) > 0)
      {
         playerItem = new PlayerItem(itemData, x, y, z, direction);

         // No se pueden llevar elementos a otras salas.
         if (hasItem)
         {
            gameManager->setItemTaken(0);
         }

         // Asigna las vidas.
         playerItem->setLives(gameManager->getLives(newPlayerId));

         // Asigna la posesión de sus objetos.
         playerItem->setTools(gameManager->hasTool(newPlayerId));

         // Asigna la cantidad de munición disponible.
         playerItem->setAmmo(gameManager->getDonuts(newPlayerId));

         // Asigna (si hay) la cantidad de grandes saltos disponibles.
         playerItem->setHighJumps(gameManager->getHighJumps());

         // Asigna (si hay) el tiempo restante de doble velocidad.
         playerItem->setHighSpeed(gameManager->getHighSpeed());

         // Asigna (si hay) el tiempo restante de inmunidad.
         playerItem->setShieldTime(gameManager->getShield(newPlayerId));

         // Un jugador necesita acceso a los datos de otros elementos dado que los
         // necesita para modelar su comportamiento. Por ejemplo: el disparo y el elemento
         // de transición entre los telepuertos.
         playerItem->assignBehavior(newBehaviorId, reinterpret_cast<void*>(itemDataManager));

         // Añade el jugador a la sala.
         room->addComponent(playerItem);
      }
   }

   return playerItem;
}

FloorTile* RoomBuilder::buildFloorTile(const boost::property_tree::ptree& tile)
{
   FloorTile *floorTile = 0;

   try
   {
      BITMAP *bitmap = load_bitmap((isomot::sharePath() + tile.get<std::string>("bitmap")).c_str(), 0);
      if (bitmap == nullptr)
      {
         throw "Bitmap " + tile.get<std::string>("bitmap") + " not found.";
      }
      int column = room->getTilesX() * tile.get<int>("y") + tile.get<int>("x");
      floorTile = new FloorTile(column, tile.get<int>("x"), tile.get<int>("y"), bitmap);
      floorTile->setOffset(tile.get<int>("offsetX"), tile.get<int>("offsetY"));
   }
   catch (std::runtime_error& e)
   {
      std::cout << e.what() << std::endl;
   }

   return floorTile;
}

Wall* RoomBuilder::buildWall(const boost::property_tree::ptree& wall)
{
   Wall* roomWall{};

   try
   {
      BITMAP* bitmap = load_bitmap((isomot::sharePath() + wall.get<std::string>("bitmap")).c_str(), 0);
      if (bitmap == nullptr)
      {
         throw "Bitmap " + wall.get<std::string>("bitmap") + " not found.";
      }
      roomWall = new Wall(wall.get<std::string>("axis") == "x" ? AxisX : AxisY,
                          wall.get<int>("index"), bitmap);
   }
   catch (std::runtime_error& e)
   {
      std::cout << e.what() << std::endl;
   }

   return roomWall;
}

GridItem* RoomBuilder::buildGridItem(const boost::property_tree::ptree& item)
{
   GridItem* gridItem{};

   // Se buscan los datos del elemento.
   ItemData* itemData = this->itemDataManager->find(item.get<short>("label"));

   // Si se han encontrado, se coloca el elemento en la sala.
   if (itemData != nullptr)
   {
      // Hay diferencia entre los enumerados de dirección manejados por el archivo y por las salas.
      std::string direction{item.get<std::string>("direction")};
      gridItem = new GridItem(itemData,
                              item.get<int>("<xmlattr>.x"),
                              item.get<int>("<xmlattr>.y"),
                              item.get<int>("<xmlattr>.z") > Top ? item.get<int>("<xmlattr>.z") * LayerHeight : Top,
                              direction == "none" ? NoDirection : Direction(directionMap[direction]));

      // Datos extra necesarios para el comportamiento del elemento.
      int behavior{item.get<int>("behavior")};
      if (behavior == VolatileTimeBehavior ||
          behavior == VolatileTouchBehavior ||
          behavior == VolatileWeightBehavior ||
          behavior == VolatileHeavyBehavior ||
          behavior == VolatilePuppyBehavior)
      {
         gridItem->assignBehavior(BehaviorId(behavior),
                                  reinterpret_cast<void*>(this->itemDataManager->find(BubblesLabel)));
      }
      else
      {
         gridItem->assignBehavior(BehaviorId(behavior), 0);
      }
   }

   return gridItem;
}

FreeItem* RoomBuilder::buildFreeItem(const boost::property_tree::ptree& item)
{
   FreeItem* freeItem{};
   int foundData{};
   int* data{};

   // Se buscan los datos del elemento.
   ItemData* itemData = this->itemDataManager->find(item.get<short>("label"));

   // Si se han encontrado, se coloca el elemento en la sala.
   if (itemData != nullptr)
   {
      // Coordenadas libres.
      int fx = item.get<int>("<xmlattr>.x") * room->getTileSize() + ((room->getTileSize() - itemData->widthX) >> 1);
      int fy = (item.get<int>("<xmlattr>.y") + 1) * room->getTileSize() - ((room->getTileSize() - itemData->widthY) >> 1) - 1;
      int fz = item.get<int>("<xmlattr>.z") != Top ? item.get<int>("<xmlattr>.z") * LayerHeight : Top;

      // La única excepción para colocar un elemento es que sea un bonus y ya se haya cogido.
      if (BonusManager::getInstance()->isPresent(room->getIdentifier(), itemData->label))
      {
         // Hay diferencia entre los enumerados de dirección manejados por el archivo y por las salas.
         std::string direction{item.get<std::string>("direction")};
         freeItem = new FreeItem(itemData, fx, fy, fz,
                                 direction == "none" ? NoDirection : Direction(directionMap[direction]));

         // Datos extra necesarios para el comportamiento del elemento.
         switch (item.get<int>("behavior"))
         {
            case ElevatorBehavior:
            {
               data = new int[3];

               // Extrae los datos extra asociados al ascensor.
               auto&& extraIter = item.equal_range("extra");
               for (auto&& i = extraIter.first; i != extraIter.second; ++i)
               {
                  data[foundData++] = boost::lexical_cast<int>(i->second.data());
               }

               // Si los tres datos necesarios constan entonces se asigna el comportamiento.
               if (foundData == 3)
               {
                  freeItem->assignBehavior(BehaviorId(item.get<int>("behavior")),
                                           reinterpret_cast<void*>(data));
                  delete[] data;
               }
               break;
            }

            case HunterWaiting4Behavior:
               freeItem->assignBehavior(BehaviorId(item.get<int>("behavior")),
                                        reinterpret_cast<void*>(this->itemDataManager->find(ImperialGuardLabel)));
               break;

            case SpecialBehavior:
            case VolatileTimeBehavior:
            case VolatileTouchBehavior:
            case VolatileWeightBehavior:
            case CannonBallBehavior:
               freeItem->assignBehavior(BehaviorId(item.get<int>("behavior")),
                                        reinterpret_cast<void*>(this->itemDataManager->find(BubblesLabel)));
               break;

            default:
               freeItem->assignBehavior(BehaviorId(item.get<int>("behavior")), 0);
         }
      }
   }

   return freeItem;
}

Door* RoomBuilder::buildDoor(const boost::property_tree::ptree& item)
{
   return new Door(this->itemDataManager,
                   item.get<short>("label"),
                   item.get<int>("<xmlattr>.x"),
                   item.get<int>("<xmlattr>.y"),
                   item.get<int>("<xmlattr>.z") > Top ? item.get<int>("<xmlattr>.z") * LayerHeight : Top,
                   Direction(directionMap[item.get<std::string>("direction")]));
}

}
